/*
** The netatopd daemon can be started optionally to read the counters
** of exited processes from the netatop module. The 'usual' state of this
** daemon is the getsockopt command NETATOP_GETCNT_EXIT. With this command
** netatopd blocks until an exited task is available on the exitlist of the 
** netatop module. The obtained netpertask struct is written to a logfile
** (similar to the process accounting info provided by the base kernel
** itself) when there is at least 5% of free space in the filesystem of
** the logfile.
**
** The logfile contains a small header struct that contains a.o. a sequence
** number indicating how many netpertask structs are currently logged.
** This header is mmapped and can be consulted by analysis processes that
** consult the logfile (e.g. atop).
** Behind the header, the netpertask structs can be found in compressed
** form. Every compressed netpertask struct is preceeded by one byte 
** specifying the size of the compressed netpertask struct.
**
** Before an analysis process starts using the logfile, it has to subscribe
** itself by decrementing the second semaphore of a semaphore group owned
** by netatopd. In this way netatopd knows how many analysis processes
** are using the logfile. When no processes use the logfile any more,
** netatopd removes the logfile and start building a new one as soon as
** a new subscription of an analysis process is noticed.
** ----------------------------------------------------------------------
** Copyright (C) 2012    Gerlof Langeveld (gerlof.langeveld@atoptool.nl)
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <errno.h>
#include <zlib.h>
#include <time.h>
#include <signal.h>
#include <sys/mman.h>
#include <sys/statvfs.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>

#include "../netatop.h"
#include "../netatopd.h"
#include "../netatopversion.h"

#include <stdio.h>
#include <string.h>

/*
** Semaphore-handling
**
** A semaphore-group with two semaphores is created. The first semaphore
** specifies the number of netatopd processes running (to be sure that only
** one daemon is active at the time) ) and the second reflects the number
** of processes using the log-file (inverted).
** This second semaphore is initialized at some high value and is
** decremented by every analysis process (like atop) that uses the log-file
** and incremented as soon as such analysis process stops again.
*/
#define SEMTOTAL	100
static int numclients(int semid)
{
	return SEMTOTAL - semctl(semid, 1, GETVAL, 0);
}

#define err_log(s)	syslog(LOG_ERR, s)
#define info_log(s)	syslog(LOG_INFO, s)

/*
** open history file
*/
static int histopen(struct naheader **nahp)
{
	int fd;
	struct naheader nahdr;

	/*
	** remove the old file; this way atop can detect that a
	** new file must be opened
	*/
	unlink(NETEXITFILE);

	/*
	** open new file
	*/
	fd = open(NETEXITFILE, O_RDWR|O_CREAT|O_TRUNC, 0644);
	if (fd == -1) {
		err_log("cannot open " NETEXITFILE " for write\n");
		return -1;
	}

	/*
	** write new header and mmap
	*/
	nahdr.magic = MYMAGIC;
	nahdr.curseq = 0;
	nahdr.hdrlen = sizeof(struct naheader);
	nahdr.ntplen = sizeof(struct netpertask);
	nahdr.mypid = getpid();
	if (write(fd, &nahdr, sizeof(nahdr)) != sizeof(nahdr)) {
		err_log("cannot write to " NETEXITFILE "\n");
		return -1;
	}

	*nahp = mmap(NULL, sizeof(*nahp), PROT_WRITE, MAP_SHARED, fd, 0);

	if (*nahp == (void *)-1) {
		err_log("mmap of " NETEXITFILE " failed\n");
		return -1;
	}

	return fd;
}

static int recstore(int fd, struct netpertask *np, socklen_t len)
{
	Byte compbuf[sizeof(*np) + 128];
	unsigned long complen;
	struct statvfs statvfs;

	/*
	** check if the filesystem is not filled for more than 95%
	*/
	if (fstatvfs(fd, &statvfs) != -1 &&
	    statvfs.f_bfree * 100 / statvfs.f_blocks < 5) {
		err_log("Filesystem > 95%% full; write skipped\n");
		return 0;
	}

	/*
	** filesystem space sufficient
	** compress netpertask struct
	*/
	complen = sizeof(compbuf) - 1;
	switch (compress(compbuf + 1, &complen, (Byte *)np, (unsigned long)sizeof(*np))) {
	case Z_OK:
	case Z_STREAM_END:
	case Z_NEED_DICT:
		break;
	default:
		err_log("compression failure\n");
		return 5;
	}

	compbuf[0] = (Byte)complen;

	/*
	** write compressed netpertask struct, headed by one byte
	** with the size of the compressed struct
	*/
	if (write(fd, compbuf, complen + 1) < complen) {
		err_log("write failure\n");
		return 5;
	}

	return 0;
}


/*
** dummy handler for SIGHUP
*/
void gethup(int sig)
{
}

int main(int argc, char *argv[])
{
	int i, netsock, histfd, semid;
	struct netpertask npt;
	socklen_t len;
	struct naheader *nap;
	struct sigaction sigact;
	struct sembuf semincr;

	/*
	** version number required?
	*/
	if (argc == 2 && *argv[1] == '-' && *(argv[1] + 1) == 'v') {
		puts(NETATOPVERSION " - " NETATOPDATE "     <gerlof.langeveld@atoptool.nl>");
		return 0;
	}

	/*
	** verify if we are running with the right privileges
	*/
	if (geteuid() != 0) {
		fputs("Root privileges are needed!\n", stderr);
		return 1;
	}

	/*
	** open socket to IP layer 
	*/
	netsock = socket(PF_INET, SOCK_RAW, IPPROTO_RAW);
	if (netsock == -1) {
		perror("open raw socket");
		return 2;
	}

	/*
	** create the semaphore group and initialize it;
	** if it already exists, verify if a netatopd daemon
	** is already running
	*/
	semid = semget(SEMAKEY, 0, 0);
	if (semid >= 0) {	// exists?
		if (semctl(semid, 0, GETVAL, 0) == 1) {
			fputs("Another netatopd is already running!", stderr);
			return 3;
		}
	} else {
		semid = semget(SEMAKEY, 2, 0600|IPC_CREAT|IPC_EXCL);
		if (semid >= 0) {
			semctl(semid, 0, SETVAL, 0);
			semctl(semid, 1, SETVAL, SEMTOTAL);
		} else {
			perror("cannot create semaphore");
			return 3;
		}
	}

	/*
	** daemonize this process
	*/
	if (fork())
		return 0;	// implicitly switch to background

	setsid();

	if ( fork() )
		return 0;

	for (i = 0; i < 1024; i++)
		if (i != netsock)
			close(i);

	umask(022);

	chdir("/tmp");

	/*
	** open syslog interface for failure messages
	*/
	openlog("netatopd", LOG_PID, LOG_DAEMON);
	info_log("version " NETATOPVERSION " actived\n");

	/*
	** raise semaphore to define a busy netatopd
	*/
	semincr.sem_num = 0;
	semincr.sem_op = +1;
	semincr.sem_flg = SEM_UNDO;
	if (semop(semid, &semincr, 1) == -1) {
		err_log("cannot increment semaphore\n");
		return 3;
	}

	/*
	** the daemon can be woken up from getsockopt by receiving 
	** the sighup signal to verify if there are no clients any more
	** (truncate exitfile)
	*/
	memset(&sigact, 0, sizeof(sigact));
	sigact.sa_handler = gethup;
	sigaction(SIGHUP, &sigact, NULL);

	/*
	** raise priority
	*/
	(void) nice(-39);

	/*
	** open history file
	*/
	histfd = histopen(&nap);
	if (histfd == -1)
		return 3;

	/*
	** continuously obtain the info about exited processes
	*/
	for (len = sizeof(npt);;)
	{
		/*
		** check if anybody is interested in the exitfile
		** if not, close and truncate it
		*/
		if (!numclients(semid) && nap->curseq)
		{
			/*
			** destroy and reopen history file
			*/
			munmap(nap, sizeof(struct naheader));
			close(histfd);
			info_log("reopen history file\n");
			histfd = histopen(&nap);
			if (histfd == -1)
				return 3;
		}

		/*
		** get next exited process (call blocks till available)
		*/
		if (getsockopt(netsock, SOL_IP, NETATOP_GETCNT_EXIT, &npt, &len))
			// getsockopt failed
			switch (errno) {
			// no netatop module loaded?
			case ENOPROTOOPT:
				sleep(10);
				continue;
			// signal received?
			case EINTR:
				continue;
			default:
				err_log("getsockopt failed\n");
				return 0;
			}
		else if (numclients(semid)) {
			// skip if nobody is using it
			/*
			** at least one listener is active, so
			** store record for exited process
			*/
			i = recstore(histfd, &npt, len);
			if (i)
				return i;
			/*
			** increment sequence number in file header
			*/
			nap->curseq++;
		}
	}

	return 0;
}

